(***************************************************************************
 *
 * Expression Tree Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   7 March 2002
 * Date:   11 October 2009
 *
 ***************************************************************************)

structure ExpTree : EXP_TREE =
  struct

    type iSet = IndexSetSeq.iSet
    type iSetSeq = IndexSetSeq.iSetSeq

    (* datatype arithOp = Add | Sub | Mul | Div
     *
     *   Arithmetic operators.
     *)
    datatype arithOp = Add | Sub | Mul | Div

    (* datatype expTree = Array of string * (string * int) list
     * 			| Fun of string * (string * int) list
     *			| IntLit of IntInf.int
     *			| Output of expTree
     * 			| Cast of (string * int) list * expTree
     * 			| Sum of string list * expTree
     * 			| BinOp of arithOp * expTree * expTree
     *
     * Ex:  Sum (["j"],
     *	      BinOp (Mul, Sum (["i"],
     *			    Array ("A", [("i",500), ("j",100)])),
     *		     Sum (["l"],
     *			  BinOp (Mul,
     *				 Array ("B", [("j",100),("k",40),("l",15)]),
     *				 Array ("C", [("k",40), ("l",15)])))))
     *
     * Array (a, i)
     *   constructs an array with array name a and a list of string*int pairs,
     *   each with a parameter name and the index range
     * Fun (f, i)
     *   constructs a function call with name f and a list of string*int pairs,
     *   each with a parameter name and the index range
     * IntLit i
     *   constructs an integer literal expression with integer i
     * Output l
     *   constructs an output expression; must be root node of the tree
     * Cast (i, l)
     *   construct an expression to cast to an array type with indices i;
     *   is typically the root node of the tree when initializing an array
     * Sum (i, l)
     *   constructs a summation expression with summation indices i and
     *   subtree l
     * BinOp (O, l, r)
     *   constructs a binary expression with operator O and  subtrees l and r
     *)
    datatype expTree = Array of string * (string * int) list
		     | Fun of string * (string * int) list
		     | IntLit of IntInf.int
		     | Output of expTree
		     | Cast of (string * int) list * expTree
		     | Sum of string list * expTree
		     | BinOp of arithOp * expTree * expTree

    local
	structure StringIntElem =
	  struct
	    type t = string * int
	    val eq = op=
	    val toString = fn x => ""
	  end

	structure StringIntSet = Set (structure Elem = StringIntElem)

      (* fusible' t
       *   returns the list of fusible indices of t with their index ranges,
       *   for efficiency, this information should be stored in the tree nodes
       *)
      fun fusible' (Array (a, i)) = i
	| fusible' (Fun (f, i)) = i
	| fusible' (IntLit i) = nil
	| fusible' (Output _) = nil
	| fusible' (Cast (i, _)) = i
	| fusible' (Sum (i, l)) =
	    List.filter (fn x => not (IndexSet.elem (#1 x, i))) (fusible' l)
	| fusible' (BinOp (_, l, r)) =
	    StringIntSet.union (fusible' l, fusible' r)

      (* sort s
       *   uses Quicksort to sort the string list s
       *   copied from IndexSet, should be in its own structure
       *)
      fun sort nil = nil
	| sort (h::t) =
	    sort (List.filter (fn x => String.< (x, h)) t)
	      @ [h] @ sort (List.filter (fn x => String.< (h, x)) t)

      fun fusible'' t = map (fn x => #1 x) (fusible' t)

    in
      (* fusible t
       *   returns the set of indices of t that are fusible with the parent
       *)
      fun fusible t = sort (fusible'' t)

      (* indices t
       *   returns the set of indices at a node
       *)
      fun indices (t as Output l) = fusible l
	| indices (t as Sum (i, l)) = fusible l
	| indices t = fusible t

      (* indexList t
       *   returns the list of indices at a node
       *)
      fun indexList (t as Output l) = fusible'' l
	| indexList (t as Sum (i, l)) = fusible'' l
	| indexList t = fusible'' t

      (* fusedSize (v, f)
       *   returns the memory needed for node v under fusion f
       *)
      fun fusedSize (Array (_, _), _) = IntInf.fromInt 0
	| fusedSize (IntLit _, _) = IntInf.fromInt 0
	| fusedSize (Output _, _) = IntInf.fromInt 0
	| fusedSize (v, f) =
	    let
	      val one = IntInf.fromInt 1
	      val i = IndexSetSeq.set f
	    in
	      foldr IntInf.* one
		    (map (fn x => if IndexSet.elem (#1 x, i) then one
				  else IntInf.fromInt (#2 x))
		         (fusible' v))
	    end

    end

    (* recomputationCost (v, f)
     *   Returns the cost for recomputations of node v under fusion f.
     *)
    fun recomputationCost (IntLit _, _) = IntInf.fromInt 0
      | recomputationCost (Array _, _) = IntInf.fromInt 0
      | recomputationCost (Output _, _) = IntInf.fromInt 0
      | recomputationCost (v, f) =
	  let
	    val one = IntInf.fromInt 1
	    val ri = IndexSet.difference (IndexSetSeq.set f, indices v)
	  in
	    IntInf.* (
	      IntInf.- (foldr IntInf.* one
			      (map (IntInf.fromInt o SymTab.getRange) ri),
		        one),
	      (foldr IntInf.* one
		     (map (IntInf.fromInt o SymTab.getRange) (indices v))))
	  end

    local 
      fun toStringOp Add = "+"
        | toStringOp Sub = "-"
        | toStringOp Mul = "*"
        | toStringOp Div = "/"

    in
      local
	(* returns a string of n spaces *)
	fun indent n = if n = 0 then "" else " " ^ indent (n - 1)

	(* constructs a string out of an (index, range) list *)
	fun indexListToString nil = ""
	  | indexListToString [(n, s)] = n ^ ":" ^ Int.toString s
	  | indexListToString ((n, s)::t) =
	      n ^ ":" ^ Int.toString s ^ ", " ^ indexListToString t

	(* expTreeToString' (t, n, m)
	 *   constructs a string out of an expTree t indented by m spaces,
	 *   with the first line only indented by n spaces
	 *)   
	fun toString' (Array (a, i), n, m) =
	      indent n ^ a ^ "[" ^ indexListToString i ^ "]\n"
	  | toString' (Fun (f, i), n, m) =
	      indent n ^ f ^ "(" ^ indexListToString i ^ ")\n"
	  | toString' (IntLit i, n, m) =
	      indent n ^ IntInf.toString i ^ "\n"
	  | toString' (Output l, n, m) =
	      indent n ^ "output\n" ^ toString' (l, m + 2, m + 2)
	  | toString' (Cast (i, l), n, m) =
	      indent n ^ "(" ^ IndexSet.toString (map #1 i) ^ ")\n"
		^ toString' (l, m + 2, m + 2)
	  | toString' (Sum (i, l), n, m) =
	      indent n ^ "sum(" ^ IndexSet.toString i ^ ")\n"
		^ toString' (l, m + 2, m + 2)
	  | toString' (BinOp (O, l, r), n, m) =
	      indent n ^ toStringOp O ^ " "
		^ toString' (l, 0, m + 2) ^ toString' (r, m+2, m + 2)
      in
	(* toString t
	 *   constructs a string out of an expTree t
	 *)
	fun toString t = toString' (t, 0, 0)
      end

      local
	(* constructs a string of indices out of an (index, range) list *)
	fun indexListToExp nil = ""
	  | indexListToExp [(n, _)] = n
	  | indexListToExp ((n, _)::t) = n ^ "," ^ indexListToExp t

      in
	(* toExp t
	 *   constructs a string in expression notation out of an expTree t
	 *)
	fun toExp (Array (a, i)) = a ^ "[" ^ indexListToExp i ^ "]"
	  | toExp (Fun (f, i)) = f ^ "(" ^ indexListToExp i ^ ")"
	  | toExp (IntLit i) = IntInf.toString i
	  | toExp (Output l) = "output " ^ toExp l
	  | toExp (Cast (i, l)) =
	      "(" ^ IndexSet.toString (map #1 i) ^ ") " ^ toExp l
	  | toExp (Sum (i, l)) =
	      "sum_" ^ IndexSet.toString i ^ " (" ^ toExp l ^ ")"
	  | toExp (BinOp (O, l, r)) =
	      toExp l ^ " " ^ toStringOp O ^ " " ^ toExp r
      end
    end

  end
