(***************************************************************************
 *
 * Fusion Tree Interface
 *
 * Author: Gerald Baumgartner
 * Date:   25 September 2001
 * Date:   9 April 2010
 *
 ***************************************************************************)

structure FusionTree : FUSION_TREE =
  struct

    type iSetSeq = SolTree.iSetSeq
    type expTree = SolTree.expTree
    type solTree = SolTree.solTree
    type solTreeSet = SolTree.solTreeSet

    (* datatype fusionTree =
     *	     Leaf   of iSetSeq * iSetSeq *
     *		       (IntInf.int * IntInf.int * IntInf.int) *
     *		       expTree
     *	   | Unary  of iSetSeq * iSetSeq *
     *		       (IntInf.int * IntInf.int * IntInf.int) *
     *		       expTree * fusionTree
     *	   | Binary of iSetSeq * iSetSeq *
     *		       (IntInf.int * IntInf.int * IntInf.int) *
     *		       expTree * fusionTree * fusionTree
     *
     * Leaf (f, h, c, t)
     *   constructs a leaf node with fusion f, nesting h, cost c, and
     *   expression tree t
     * Unary (f, h, c, t, F1)
     *   constructs a unary node with fusion f, nesting h, cost c,
     *   expression tree t, and subtree F1
     * Binary (f, h, c, t, F1, F2)
     *   constructs a binary node with fusion f, nesting h, cost c,
     *   expression tree t, and subtrees F1 and F2
     *
     * where c = (hi, lo, rc) consists of
     *   hi	the memory cost for a static allocation cost model
     *		the maximum memory cost for a dynamic allocation cost model
     *   lo	the cost for holding the result of a tree for dynamic model
     *   rc	the recomputation cost for space-time tradeoffs
     *)
    datatype fusionTree =
	     Leaf   of iSetSeq * iSetSeq *
		       (IntInf.int * IntInf.int * IntInf.int) *
		       expTree
	   | Unary  of iSetSeq * iSetSeq *
		       (IntInf.int * IntInf.int * IntInf.int) *
		       expTree * fusionTree
	   | Binary of iSetSeq * iSetSeq *
		       (IntInf.int * IntInf.int * IntInf.int) *
		       expTree * fusionTree * fusionTree

    local
      (* returns a string of n spaces *)
      fun indent n = if n = 0 then "" else " " ^ indent (n - 1)

      (* produce the string for a single node *)
      fun toString'' (f, h, (hi, lo, rc), t, n) =
	indent n ^ "(" ^ IndexSetSeq.toString f ^ ", "
	  ^ IndexSetSeq.toString h ^ "), ("
	  ^ IntInf.toString hi ^ "," ^ IntInf.toString lo ^ ","
	  ^ IntInf.toString rc ^ "), "
	  ^ ExpTree.toExp t ^ "\n"

      fun toString' (Leaf (f, h, c, t), n) = toString'' (f, h, c, t, n)
	| toString' (Unary (f, h, c, t, l), n) =
	    toString'' (f, h, c, t, n) ^ toString' (l, n + 2)
	| toString' (Binary (f, h, c, t, l, r), n) =
	    toString'' (f, h, c, t, n)
	      ^ toString' (l, n + 2) ^ toString' (r, n + 2)

    in
      (* toString F
       *   convert the fusion tree F into a string
       *)
      fun toString F = toString' (F, 0)
    end

    local
      fun C ((hi, lo, _, _), rc) = (hi, lo, rc)

      (* skip Reduced solutions nodes and set the cost *)
      fun skipReducedSolns (SolTree.Reduced (_, _, l), f, c) =
	    skipReducedSolns (l, f, c)
	| skipReducedSolns (t, f, c) = buildFusionTree'' (t, f, c)
      and
      (* set the cost from the enclosing Extended/Reduced node *)
          buildFusionTree'' (SolTree.Leaf (Solution.S (h, _, _, _, t)), f, c) =
	    Leaf (f, IndexSetSeq.suffix (h, f), c, t)
	| buildFusionTree'' (SolTree.Extended (Solution.S (f',_,_,_,_),_,l),f,c) =
	    buildFusionTree'' (l, IndexSetSeq.restrict (f', f), c)
	| buildFusionTree'' (SolTree.Reduced (_, _, l), f, c) =
	    skipReducedSolns (l, f, c)
	| buildFusionTree'' (SolTree.Unary (Solution.S (h,_,_,_,t), l), f, c) =
	    Unary (f, IndexSetSeq.suffix (h, f), c, t,
		   buildFusionTree' (l, IndexSetSeq.restrict (h, f)))
	| buildFusionTree'' (SolTree.Binary (Solution.S (h,_,_,_,t), l, r), f, c) =
	    Binary (f, IndexSetSeq.suffix (h, f), c, t,
		    buildFusionTree' (l, IndexSetSeq.restrict (h, f)),
		    buildFusionTree' (r, IndexSetSeq.restrict (h, f)))
      and
      (* propagate the fusion down to the leaf nodes and restrict them *)
          buildFusionTree' (SolTree.Leaf (Solution.S (h, c, _, rc, t)), f) =
	    Leaf (f, IndexSetSeq.suffix (h, f), C(c,rc), t)
	| buildFusionTree' (SolTree.Extended (Solution.S (f',c,_,rc,_), _, l), f) =
	    buildFusionTree'' (l, IndexSetSeq.restrict (f', f), C(c,rc))
	| buildFusionTree' (SolTree.Reduced (Solution.S (_,c,_,rc,_), _, l), f) =
	    skipReducedSolns (l, f, C(c,rc))
	| buildFusionTree' (SolTree.Unary (Solution.S (h,c,_,rc,t), l), f) =
	    Unary (f, IndexSetSeq.suffix (h, f), C(c,rc), t,
		   buildFusionTree' (l, IndexSetSeq.restrict (h, f)))
	| buildFusionTree' (SolTree.Binary (Solution.S (h,c,_,rc,t), l, r), f) =
	    Binary (f, IndexSetSeq.suffix (h, f), C(c,rc), t,
		    buildFusionTree' (l, IndexSetSeq.restrict (h, f)),
		    buildFusionTree' (r, IndexSetSeq.restrict (h, f)))
    in
      (* buildFusionTree S
       *   convert the solution tree S into a fusion tree by propagating
       *   the fusion down from the root while skipping reduced and
       *   extended solutions
       *)
      fun buildFusionTree S = buildFusionTree' (S, nil)
    end

    (* expTreeToFusionTree T
     *   runs the minMemFusion algorithm on expression tree T and
     *   converts the result into a fusion tree
     *)
    fun expTreeToFusionTree T = buildFusionTree (SolTree.minMemFusion T)

    local
      val zero = IntInf.fromInt 0

      fun buildUnfusedFusionTree' (t as ExpTree.Array _, i, _) =
	    Leaf ([i], [], (zero, zero, zero), t)
        | buildUnfusedFusionTree' (t as ExpTree.Fun _, _, _) =
	    Leaf ([], [ExpTree.indices t], (zero, zero, zero), t)
        | buildUnfusedFusionTree' (t as ExpTree.IntLit _, i, _) =
	    Leaf ([i], [], (zero, zero, zero), t)
        | buildUnfusedFusionTree' (t as ExpTree.Output l, _, _) =
	    Unary ([], [ExpTree.indices t], (zero, zero, zero), t,
		   buildUnfusedFusionTree' (l, ExpTree.indices t, true))
        | buildUnfusedFusionTree' (t as ExpTree.Cast (_, l), _, _) =
	    Unary ([], [ExpTree.indices t], (zero, zero, zero), t,
		   buildUnfusedFusionTree' (l, ExpTree.indices t, true))
        | buildUnfusedFusionTree' (t as ExpTree.Sum (_, l), _, _) =
	    Unary ([], [ExpTree.indices t], (zero, zero, zero), t,
		   buildUnfusedFusionTree' (l, ExpTree.indices t, true))
        | buildUnfusedFusionTree' (t as ExpTree.BinOp (_, l, r), _, u) =
	    if u then
	      Binary ([ExpTree.indices t], [], (zero, zero, zero), t,
		      buildUnfusedFusionTree' (l, ExpTree.indices t, false),
		      buildUnfusedFusionTree' (r, ExpTree.indices t, false))
	    else
	      Binary ([], [ExpTree.indices t], (zero, zero, zero), t,
		      buildUnfusedFusionTree' (l, ExpTree.indices t, false),
		      buildUnfusedFusionTree' (r, ExpTree.indices t, false))
    in
      (* buildUnfusedFusionTree T
       *   builds s a fusion tree without fusions directly from the
       *   expression tree T
       *)
      fun buildUnfusedFusionTree t = buildUnfusedFusionTree' (t, [], false)
    end
  end
