(***************************************************************************
 *
 * Set Implementation
 *
 * Author: Gerald Baumgartner
 * Date:   27 October 2001
 *
 ***************************************************************************)

functor Set (structure Elem : ELEM) : SET =
  struct
    (* t
     *   the element type
     *)
    type t = Elem.t

    (* 'a set
     *    the set representation
     *    Ex: [1, 2, 3]
     *)
    type 'a set = 'a list

    (* empty
     *   the empty set
     *)
    val empty = nil

    (* isEmpty s
     *   returns true iff s is empty
     *)
    fun isEmpty s = s = nil

    (* elem (e, s)
     *   returns true iff element e is in set s
     *   Ex: elem (1, [1, 2, 3]) returns true
     *)
    fun elem (e, s) = List.exists (fn x => Elem.eq (x, e)) s

    (* subset (s1, s2)
     *   returns true iff set s1 is a subset of set s2
     *   Ex: subset ([1, 2], [1, 2, 3]) returns true
     *)
    fun subset (s1, s2) = List.all (fn x => elem (x, s2)) s1

    (* eq (s1, s2)
     *   returns true iff set s1 and set s2 contain the same elements
     *   Ex: eq ([1, 2], [2, 1]) returns true
     *)
    fun eq (s1, s2) = subset (s1, s2) andalso subset (s2, s1)

    (* union (s1, s2)
     *   returns the union of sets s1 and s2
     *   Ex: union ([1, 2], [2, 3]) returns [3, 1, 2]
     *)
    fun union (l, nil) = l
      | union (l, (h::t)) =
	  if List.exists (fn x => Elem.eq (x, h)) l then union (l, t)
	  else h :: union (l, t)

    (* intersect (s1, s2)
     *   returns the intersection of sets s1 and s2
     *   Ex: intersect ([1, 2], [2, 3]) returns [2]
     *)
    fun intersect (l, nil) = nil
      | intersect (l, (h::t)) =
	  if List.exists (fn x => Elem.eq (x, h)) l then h :: intersect (l, t)
	  else intersect (l, t)

    (* Union S
     *   returns the union of the sets in list S
     *   Ex: Union [[1, 2], [2, 3]] returns [3, 1, 2]
     *)
    fun Union nil = nil
      | Union (h::t) = union (h, Union t)

    (* Intersect S
     *   returns the intersection of the sets in list S
     *   Ex: Intersect [[1, 2], [2, 3]] returns [2]
     *)
    fun Intersect nil = nil
      | Intersect [h] = h
      | Intersect (h::t) = intersect (h, Intersect t)

    (* difference (s1, s2)
     *   returns the set of all the elements from s1 that are not in s2
     *   Ex: difference ([1, 2, 3], [2, 3]) returns [1]
     *)
    fun difference (s1, s2) =
      List.filter (fn x => not (elem (x, s2))) s1

    (* powerSet s
     *   returns the set of subsets of s
     *)
    fun powerSet nil = [nil]
      | powerSet (h::t) =
	  let val p = powerSet t
	  in
	    p @ (map (fn s => h :: s) p)
	  end

    (* crossProduct (S1, S2)
     *   returns the cross product of sets S1 and S2
     *)
    fun crossProduct (S1, S2) =
      foldr (fn (x,p) => foldr (fn (y,q) => (x,y)::q) p S2) [] S1

    (* setMap f S
     *   map the function f onto set S
     *)
    val setMap = map

    local
      (* returns a string representation of set s without curly braces *)
      fun toString' nil = ""
	| toString' [h] = Elem.toString h
	| toString' (h::t) = concat [Elem.toString h, ", ", toString' t]
    in
      (* toString s
       *   returns a string representation of set s
       *)
      fun toString s = concat ["{", toString' s, "}"]
    end
  end
